import Fraction from "fraction.js";
import cuid from "cuid";
import x from "xastscript";
import Duration from "./duration";
import setDuration from "./util/setDuration";
import Pitch from "./pitch";
import Direction from "./direction";
import Octave from "./octave";
import distinctOrArray from "./util/distinctOrArray";
import setMusicXmlTie from "./util/musicXml/setTie";
import setType from "./util/musicXml/setType";
import setDots from "./util/musicXml/setDots";
import {
  EventProps,
  DurationInput,
  PitchInput,
  MusicXmlProps,
  MeiProps,
  EventDataProps,
} from "./types";

class Event extends Duration {
  children!: Pitch[];
  props: EventProps;

  constructor(children: Pitch[] = [], props?: EventProps) {
    if (!props) {
      props = {
        duration: setDuration(),
        type: "rest",
        tie: false,
        directions: [],
        id: "",
      };
    }

    const duration = props.duration.beats;
    const lcd = new Fraction(duration.d / 4);
    super(children, duration, lcd);
    this.props = props;
  }

  setDuration(value: DurationInput, hidden = false) {
    const duration = setDuration(value);
    const type = hidden ? "space" : this.props.type;
    return this._copy(this.children, { duration, type });
  }

  setPitches(input: PitchInput | PitchInput[]): Event;
  setPitches(...input: PitchInput[]): Event;
  setPitches(...input: any) {
    const values: PitchInput[] = distinctOrArray(input);
    const pitches: Pitch[] = [];
    for (const value of values) {
      const pitch = new Pitch().setPitch(value);
      if (pitch) {
        pitches.push(pitch);
      }
    }
    if (pitches.length) {
      const type = pitches.length === 1 ? "note" : "chord";
      return this._copy(pitches, { type });
    } else {
      return this;
    }
  }

  addDirection(value: Direction) {
    let { directions } = this.props;
    directions = [...directions, value];
    return this._copy(this.children, { directions });
  }

  setTie(tie: boolean) {
    return this._copy(this.children, { tie });
  }

  relativeToFixed(octave: Octave) {
    if (this.children.length) {
      const pitches = this.children.map((pitch, i) => {
        pitch = pitch.relativeToFixed(octave);
        if (i === 0) {
          octave = octave.clone;
        }
        return pitch;
      });
      return this._copy(pitches);
    } else {
      return this;
    }
  }

  setTies(previous: number[], current: number[], next: number[]) {
    const stop = current.map((pitchNum) => previous.includes(pitchNum));
    const start = current.map((pitchNum) =>
      this.props.tie ? next.includes(pitchNum) : false
    );
    const pitches = this.children.map((pitch, i) =>
      pitch.setTie({ stop: stop[i], start: start[i] })
    );
    return this._copy(pitches);
  }

  setId() {
    const id = cuid();
    return this._copy(this.children, { id });
  }

  private _copy(children: Pitch[], change: any = {}) {
    return new Event(children, {
      ...this.props,
      ...change,
    });
  }

  toMusicXml({ lcd, timeMod, staff, voice, tuplets }: MusicXmlProps) {
    const { type: eventType, duration, directions } = this.props;
    const { type, dots, beats } = duration;
    const notes =
      eventType === "space" || eventType === "rest"
        ? [x("rest")]
        : this.children.map((pitch) => pitch.toMusicXml());

    return [
      ...directions.map((direction) => direction.toMusicXml()),
      ...notes.map((note, i) =>
        x("note", eventType === "space" ? { "print-object": "no" } : {}, [
          i !== 0 ? x("chord") : null,
          note,
          x("duration", beats.mul(4).mul(lcd).mul(timeMod.inverse()).valueOf()),
          ...setMusicXmlTie(this.children[i]),
          x("voice", voice),
          x("type", setType(type)),
          ...setDots(dots),
          timeMod.compare(1)
            ? x("time-modification", [
                x("actual-notes", timeMod.n),
                x("normal-notes", timeMod.d),
              ])
            : null,
          x("staff", staff),
          i === 0
            ? x("notations", [
                tuplets.start.map(
                  ({ number, outer, inner, type, dots, hidden }) => {
                    const visibility = hidden
                      ? { "show-number": "none", bracket: "no" }
                      : {};
                    return x(
                      "tuplet",
                      { number, type: "start", ...visibility },
                      [
                        x("tuplet-actual", [
                          x("tuplet-number", outer),
                          x("tuplet-type", setType(type)),
                          ...setDots(dots),
                        ]),
                        x("tuplet-normal", [
                          x("tuplet-number", inner),
                          x("tuplet-type", setType(type)),
                          ...setDots(dots),
                        ]),
                      ]
                    );
                  }
                ),
                tuplets.stop.map((number) =>
                  x("tuplet", { number, type: "stop" })
                ),
              ])
            : null,
        ])
      ),
    ];
  }

  toMei(props: MeiProps) {
    const { type: eventType, duration, id, directions } = this.props;
    if (directions.length) {
      props.directions.push({ id, directions });
    }
    const { type, dots } = duration;
    const tempo = props.tempo
      ? { num: props.tempo.n, numbase: props.tempo.d }
      : {};
    const shared = {
      dur: type,
      dots,
      ...tempo,
      "xml:id": id,
    };
    switch (eventType) {
      case "space":
        return x("space", shared);
      case "rest":
        return x("rest", shared);
      case "note":
        return this.children[0].toMei(shared);
      case "chord":
        return x(
          "chord",
          shared,
          this.children.map((pitch) => pitch.toMei())
        );
    }
  }

  toEventData({ lcd, timeMod }: EventDataProps) {
    const { id } = this.props;
    return {
      id,
      duration: this.duration.mul(4).mul(lcd).mul(timeMod.inverse()).valueOf(),
      pitches: this.children.map((pitch) => pitch.toEventData()),
    };
  }
}

export default Event;
