import Fraction from "fraction.js";
import x from "xastscript";
import Layers from "./layers";
import Staff from "./staff";
import Layer from "./layer";
import setClef from "./util/musicXml/setClef";
import TempoFragment from "./tempoFragment";
import AccelFragment from "./accelFragment";
import {
  DurationInput,
  PipeProps,
  MusicXmlProps,
  MeiProps,
  EventDataProps,
} from "./types";

class Fragment extends Layers {
  children!: Staff[];

  constructor(children: Staff[] = []) {
    super(children);
    this.children = this.children.map((staff, i) =>
      staff.pipe((layer) => layer.addRests(this.duration), {
        staffNum: i,
        layerNum: 0,
      })
    );
  }

  add(value: Layer | Fragment) {
    const children = [...this.children];
    if (value instanceof Layer) {
      if (!children.length) {
        children.push(new Staff([value]));
      } else {
        children[0] = children[0].add(value);
      }
    } else if (value instanceof Fragment) {
      value.children.forEach((staff, i) => {
        if (i === children.length) {
          children.push(new Staff([new Layer().addRests(this.duration)]));
        }
        children[i] = children[i].add(staff);
      });
    } else {
      throw Error(`${value} is not a Layer or Fragment`);
    }
    return new Fragment(children);
  }

  merge(value: Fragment, staffNum: number) {
    const children = [...this.children];
    value.children.forEach((staff, i) => {
      i += staffNum;
      if (i === children.length) {
        children.push(staff);
      } else {
        children[i] = children[i].merge(staff);
      }
    });
    return new Fragment(children);
  }

  tuplet(
    outer: number,
    inner: number,
    beatValue?: DurationInput,
    hidden?: boolean
  ) {
    const children = this.children.map((staff) =>
      staff.tuplet(outer, inner, beatValue, hidden)
    );
    return new Fragment(children);
  }

  tempo(tempo: any, beatValue: DurationInput = "4") {
    tempo = new Fraction(tempo);
    return new TempoFragment(this, tempo, beatValue);
  }

  accel(
    startTempo: any,
    endTempo: any,
    numBeats: number,
    beatValue: DurationInput = "4"
  ) {
    startTempo = new Fraction(startTempo);
    endTempo = new Fraction(endTempo);
    return new AccelFragment(this, startTempo, endTempo, numBeats, beatValue);
  }

  setTies() {
    return this.pipe((layer) => layer.setTies());
  }

  transpose(interval: number) {
    return this.pipe((layer) => layer.transpose(interval));
  }

  invert(axis: number) {
    return this.pipe((layer) => layer.invert(axis));
  }

  pipe(
    ...transforms: Array<(layer: Layer, props: PipeProps) => Layer>
  ) {
    const children = transforms.reduce(
      (acc, cur) =>
        acc.map((staff, i) => staff.pipe(cur, { staffNum: i, layerNum: 0 })),
      this.children
    );
    return new Fragment(children);
  }

  toMusicXml(id = "P1", clefs: string[]) {
    const { children, lcd } = this;
    let voice = 1;
    return x("part", { id }, [
      x("measure", { number: 1 }, [
        x("attributes", [
          // REVISE DIVISIONS
          x("divisions", lcd.valueOf()),
          x("staves", children.length),
          ...clefs.map((clef, i) => setClef(clef, i + 1)),
        ]),
        ...this.children.map((staff, i) => {
          const props: MusicXmlProps = {
            lcd,
            timeMod: new Fraction(1),
            staff: i + 1,
            voice,
            tuplets: {
              number: 0,
              start: [],
              stop: [],
            },
          };
          voice += staff.children.length;
          const output = staff.toMusicXml(props);
          if (i !== this.children.length - 1) {
            output.push(
              x("backup", [
                x("duration", staff.duration.mul(4).mul(props.lcd).valueOf()),
              ])
            );
          }
          return output;
        }),
      ]),
    ]);
  }

  toMei(props: MeiProps) {
    const { staffNum } = props;
    return this.children.map((staff, i) =>
      x("staff", { n: staffNum + i }, staff.toMei(props))
    );
  }

  toEventData(props: EventDataProps) {
    return this.children.flatMap((staff) => staff.toEventData(props));
  }

  toBeatData(lcd: Fraction) {
    return this.children[0].toBeatData(lcd);
  }

  toTimingData() {
    return this.children[0].toTimingData();
  }
}

export default Fragment;
