import Fraction from "fraction.js";
import Event from "./event";
import Entries from "./entries";
import Tuplet from "./tuplet";
import Tempo from "./tempo";
import Direction from "./direction";
import Octave from "./octave";
import distinctOrArray from "./util/distinctOrArray";
import setDuration, { notPowerOfTwo } from "./util/setDuration";
import getEvents from "./util/getEvents";
import {
  Entry,
  EventInput,
  DurationInput,
  TempoProps,
  AccelProps,
  MusicXmlProps,
  MeiProps,
  EventDataProps,
} from "./types";
import { TempoObject } from "score-timing-data";

class Layer extends Entries {
  children!: Entry[];

  constructor(values: Entry[] = [], props = {}) {
    super(values, props);
  }

  static from(input: EventInput | EventInput[]): Layer;
  static from(...input: EventInput[]): Layer;
  static from(...input: any) {
    const values: EventInput[] = distinctOrArray(input);
    const layerEvents: Event[] = [];
    for (const value of values) {
      const pitches = value.pitches === undefined ? [] : value.pitches;
      const duration = value.duration || [1, 4];
      const tie = !!value.tie;
      const hidden = value.hidden || false;
      const durations: (string | number)[] =
        typeof duration === "string"
          ? [duration]
          : Layer._splitDuration(duration);
      durations.forEach((dur, i) => {
        layerEvents.push(
          new Event()
            .setPitches(pitches)
            .setDuration(dur, hidden)
            .setTie(i === durations.length - 1 ? tie : true)
        );
      });
    }
    return new Layer(layerEvents);
  }

  add(value: Event | Layer) {
    return new Layer([this, value]);
  }

  addRests(total: Fraction, hidden = true) {
    const duration = total.sub(this.duration);
    if (duration.compare(0)) {
      return this.add(Layer.from({ duration, hidden }));
    } else {
      return this;
    }
  }

  private static _splitDuration(duration: any) {
    let remaining: Fraction;
    try {
      remaining = new Fraction(duration);
    } catch (e) {
      throw Error(`Duration '${duration}' not valid`);
    }
    if (notPowerOfTwo(remaining.d)) {
      throw Error(`Duration '${duration}' must be a power of two`);
    }
    const durations: number[] = [];
    let current = new Fraction(1 / 8);
    while (remaining.compare(0)) {
      const result = remaining.sub(current);
      if (result.compare(0) >= 0) {
        durations.push(current.valueOf());
        remaining = result;
      } else {
        current = current.div(2);
      }
    }
    return durations;
  }

  tuplet(
    outer: number,
    inner: number,
    beatValue: DurationInput = "4",
    hidden = false
  ): Layer {
    const { value, type, dots, beats } = setDuration(beatValue);
    const tupletDuration = beats.mul(outer);
    let duration = tupletDuration;
    let numTuplets = 1;
    while (duration.compare(this.duration) < 0) {
      duration = duration.add(tupletDuration);
      numTuplets++;
    }
    outer *= numTuplets;
    inner *= numTuplets;
    const gap = duration.sub(this.duration).mul(beats.d);
    // console.log(`outer: ${outer} inner: ${inner} thisDuration: ${this.duration.toFraction()} duration: ${duration.toFraction()} gap: ${gap.toFraction()}`)
    const layer = gap.compare(0)
      ? this.add(
          Layer.from({ duration: beats }).tuplet(
            1,
            gap.valueOf(),
            value,
            true
          )
        )
      : this;
    const tuplet = new Tuplet([layer], {
      outer,
      inner,
      value,
      type,
      dots,
      beats,
      hidden,
    });
    // const tuplet = new Tuplet([this.addRests(duration, false)], { outer, inner, value, type, dots, beats, hidden })
    return new Layer([tuplet]);
  }

  tempo(props: TempoProps) {
    const tempo = new Tempo([this], props);
    return new Layer([tempo]);
  }

  accel(props: AccelProps) {
    const { startTempo, endTempo, numBeats, type, dots, beats } = props;
    const startDuration = startTempo.inverse();
    const endDuration = endTempo.inverse();
    const stepChange = endDuration.sub(startDuration).div(numBeats);
    const tempoArray: Fraction[] = [];
    for (let i = 0; i < numBeats; i++) {
      const duration = startDuration.add(stepChange.mul(i));
      tempoArray.push(duration.inverse());
    }
    const tempos: Tempo[] = [];
    let beatEvents: Event[] = [];
    let elapsed = new Fraction(0);
    let beatIndex = 0;
    for (const entry of this.children) {
      if (entry instanceof Event) {
        beatEvents.push(entry);
        elapsed = elapsed.add(entry.duration);
        if (elapsed.compare(beats) > 0) {
          throw Error(`Accel events cannot exceed the beat duration`);
        } else if (elapsed.equals(beats)) {
          tempos.push(
            new Tempo([new Layer(beatEvents)], {
              tempo: tempoArray[beatIndex],
              type,
              dots,
              beats,
            })
          );
          beatEvents = [];
          elapsed = new Fraction(0);
          beatIndex++;
        }
      } else {
        throw Error(`An accel can only contain events`);
      }
    }
    return new Layer(tempos);
  }

  addDirection(value: Direction) {
    let start = true;
    const transform = (event: Event) => {
      if (!start) return event;
      start = false;
      return event.addDirection(value);
    };
    return this.pipeEvents(transform);
  }

  relativeToFixed(startOctave: number) {
    const octave = new Octave(startOctave);
    const transform = (event: Event) => event.relativeToFixed(octave);
    return this.pipeEvents(transform);
  }

  setTies() {
    const setPitches = (event?: Event) =>
      event ? event.children.map((pitch) => pitch.pitchNum) : [];
    const pitches = getEvents(this.children, setPitches);
    let previous: number[] = [];
    let current = pitches.next().value;
    const transform = (entries: Entry[]) =>
      entries.map((entry) => {
        if (entry instanceof Event) {
          const next = pitches.next().value;
          entry = entry.setTies(previous, current, next);
          previous = entry.props.tie ? current : [];
          current = next;
          return entry;
        } else {
          return entry.pipe(transform);
        }
      });
    return this.pipe(transform);
  }

  transpose(interval: number) {
    const transform = (event: Event) =>
      event.setPitches(
        event.children.map(({ pitchNum }) => pitchNum + interval)
      );
    return this.pipeEvents(transform);
  }

  invert(axis: number) {
    const transform = (event: Event) =>
      event.setPitches(
        event.children.map(({ pitchNum }) => axis + (axis - pitchNum))
      );
    return this.pipeEvents(transform);
  }

  toMusicXml(props: MusicXmlProps): any {
    return this.children.map((entry) => entry.toMusicXml(props));
  }

  toMei(props: MeiProps): any {
    return this.children.map((entry) => entry.toMei(props));
  }

  toEventData(props: EventDataProps) {
    const events: any = [];
    this.children.forEach((entry) =>
      entry instanceof Event
        ? events.push(entry.toEventData(props))
        : events.push(...entry.toEventData(props))
    );
    return events;
  }

  *toBeatData(lcd: Fraction) {
    let offset = new Fraction(0);
    // yield(0)
    for (const tempo of this.children as Tempo[]) {
      yield* tempo.toBeatData(lcd, offset);
      offset = offset.add(tempo.duration);
    }
  }

  toTimingData() {
    const tempos: TempoObject[] = [];
    let position = new Fraction(0);

    for (const tempo of this.children as Tempo[]) {
      tempos.push({
        tempo: tempo.props.tempo.toFraction(),
        startPosition: position.toFraction(),
      });

      position = position.add(tempo.duration);
    }

    return {
      tempos,
      endPosition: position,
    };
  }
}

export default Layer;
