import Fragment from "./fragment";
import TempoFragment from "./tempoFragment";
import AccelFragment from "./accelFragment";
import Layer from "./layer";
import Tempo from "./tempo";
import { PartProps, MeiProps, EventDataProps } from "./types";
import Fraction from "fraction.js";

class Part {
  fragment: Fragment;
  props: PartProps;

  constructor(
    value: Fragment | (Fragment | TempoFragment | AccelFragment)[],
    name = "",
    clefs: string[] = ["treble", "bass"]
  ) {
    let fragment = new Fragment();
    if (!Array.isArray(value)) {
      value = [value];
    }
    value.forEach((tempo) => {
      if (tempo instanceof Fragment) {
        tempo = tempo.tempo(1, tempo.duration.d.toString());
      }
      if (tempo instanceof TempoFragment) {
        let { fragment: tempoFragment, props: tempoProps, direction } = tempo;
        tempoFragment = tempoFragment.pipe((layer, props) => {
          layer = layer.tempo(tempoProps);
          if (props.staffNum === 0 && props.layerNum === 0) {
            layer = layer.addDirection(direction);
          }
          return layer;
        });
        fragment = fragment.add(tempoFragment);
      } else if (tempo instanceof AccelFragment) {
        let { fragment: accelFragment, props: accelProps, direction } = tempo;
        accelFragment = accelFragment.pipe((layer, props) => {
          layer = layer.accel(accelProps);
          if (props.staffNum === 0 && props.layerNum === 0) {
            layer = layer.addDirection(direction);
          }
          return layer;
        });
        fragment = fragment.add(accelFragment);
      } else {
        throw Error(`Part input not Fragment or Tempo Fragments`);
      }
    });

    fragment = fragment.setTies();
    fragment = fragment.pipe((layer) =>
      layer.pipeEvents((event) => event.setId())
    );

    this.fragment = fragment;
    this.props = { name, clefs };
  }

  get name() {
    return this.props.name;
  }

  toMusicXml(id: string) {
    const { clefs } = this.props;
    const fragment = this.fragment.pipe((layer) => {
      const tempos = layer.children as Tempo[];
      let tuplets = tempos.map((tempo) => new Layer(tempo.children));
      const beatValue = tuplets.reduce(
        (acc, { duration }) => acc.lcm(duration.d),
        new Fraction(1)
      );
      // console.log('beatValue', beatValue)
      let ratios = tuplets.map(({ duration }, i) => {
        const { tempo } = tempos[i].props;
        const n = duration.mul(beatValue);
        const d = n.div(tempo);
        return { n, d };
      });
      // console.log('ratios', ratios)
      const lcd = ratios.reduce((acc, { d }) => acc.lcm(d.d), new Fraction(1));
      // console.log('lcd', lcd)
      tuplets = tuplets.map((tuplet, i) => {
        const { n, d } = ratios[i];
        return tuplet.tuplet(
          n.valueOf(),
          d.mul(lcd).valueOf(),
          beatValue.toString(),
          true
        );
      });
      let newLayer = new Layer(tuplets);
      let n = newLayer.duration.mul(beatValue);
      while (n.div(lcd).mod(1).compare(0)) {
        n = n.add(1);
      }
      const d = n.div(lcd);
      newLayer = newLayer.tuplet(
        n.valueOf(),
        d.valueOf(),
        beatValue.toString(),
        true
      );
      return newLayer;
    });
    return fragment.toMusicXml(id, clefs);
  }

  toMei(props: MeiProps) {
    return this.fragment.toMei(props);
  }

  toEventData(props: EventDataProps) {
    return this.fragment.toEventData(props);
  }

  toBeatData(lcd: Fraction) {
    return this.fragment.toBeatData(lcd);
  }

  toTimingData() {
    return this.fragment.toTimingData();
  }
}

export default Part;
