import x from "xastscript";
import Score from "./score";
import pitchClasses from "./util/pitchClasses";
import Octave from "./octave";
import setAlter from "./util/musicXml/setAlter";
import setMeiTie from "./util/mei/setTie";
import { Tie, PitchProps } from "./types";

class Pitch {
  props: PitchProps;

  constructor(props?: PitchProps) {
    if (!props) {
      props = {
        pitch: "",
        pitchClass: 0,
        pname: "",
        accid: "",
        oct: 0,
        tie: { stop: false, start: false },
      };
    }
    this.props = props;
  }

  get pitchNum() {
    return this.props.pitchClass + (this.props.oct + 1) * 12;
  }

  setPitch(value: string | number) {
    let pitch;
    if (typeof value === "string") {
      pitch = Pitch._fromString(value);
    } else if (typeof value === "number") {
      pitch = Pitch._fromNumber(value);
    } else {
      throw Error(`Pitch '${value}' must be string or number`);
    }
    return pitch ? this._copy(pitch) : null;
  }

  setTie(tie: Tie) {
    return this._copy({ tie });
  }

  relativeToFixed(octave: Octave) {
    let { oct, pname } = this.props;
    oct = octave.next(oct, pname);
    return this._copy({ oct });
  }

  private static _fromString(pitch: string) {
    const { pname, accid, oct } = Pitch._matchString(pitch);
    if (pname === "r") {
      if (accid) {
        throw Error(`Rest '${pitch}' should not contain accidentals`);
      }
      if (oct) {
        throw Error(`Rest '${pitch}' should not contain octave changes`);
      }
    } else {
      return {
        pitch,
        pitchClass: Pitch._setPitchClass(pname, accid),
        pname,
        accid,
        oct: Pitch._setOctave(oct),
      };
    }
  }

  private static _fromNumber(value: number) {
    const pitchClass = value % 12;
    const pitch = Score.pitchNames[pitchClass];
    const oct = (value - pitchClass - 12) / 12;
    const { pname, accid } = Pitch._matchString(pitch);
    return {
      pitch,
      pitchClass,
      pname,
      accid,
      oct,
    };
  }

  private static _matchString(value: string) {
    const match = value.match(
      /(?<pname>[a-g]|[r])(?<accid>[sf]*)(?<oct>[',]*)/
    );
    if (!match?.groups) {
      throw Error(`Pitch '${value}' is not valid`);
    }
    return match.groups;
  }

  private static _setOctave(oct: string) {
    let output = 0;
    for (const val of oct) {
      output += val === "," ? -1 : val === "'" ? 1 : 0;
    }
    return output;
  }

  private static _setPitchClass(pname: string, accid: string) {
    const pitchClass = pitchClasses.findIndex((pitches) =>
      pitches.includes(`${pname}${accid}`)
    );
    if (pitchClass === -1) {
      throw Error(`Pitch ${pname}${accid} not found`);
    }
    // -2 offsets cff and cf
    return pitchClass - 2;
  }

  private _copy(change: any = {}) {
    return new Pitch({
      ...this.props,
      ...change,
    });
  }

  toMusicXml() {
    const { pname, accid, oct } = this.props;
    return x("pitch", [
      x("step", pname.toUpperCase()),
      x("alter", setAlter(accid)),
      x("octave", oct),
    ]);
  }

  toMei(shared: any = {}) {
    const { pname, accid, oct, tie } = this.props;
    return x("note", { ...shared, pname, accid, oct, tie: setMeiTie(tie) });
  }

  toEventData() {
    const { tie } = this.props;
    return {
      num: this.pitchNum,
      tie: tie.start,
    };
  }
}

export default Pitch;
