import toXml from "xast-util-to-xml";
import u from "unist-builder";
import x from "xastscript";
import Fraction from "fraction.js";
import Fragment from "./fragment";
import Part from "./part";
import setClef from "./util/mei/setClef";
import { ScoreProps, OutputFormats, MeiDirections } from "./types";
import {
  MeterParts,
  TempoParts,
  Players,
  Player,
  ScoreTimingData,
} from "score-timing-data";

class Score {
  static octave: number | null = 4;
  private static _pitchNames = [
    "c",
    "cs",
    "d",
    "ds",
    "e",
    "f",
    "fs",
    "g",
    "gs",
    "a",
    "as",
    "b",
  ];
  parts: Part[];
  ids: string[];
  props: ScoreProps = {
    baseTempo: {
      beatUnit: "1/4",
      bpm: 120,
    },
  };

  constructor(parts: (Fragment | Part)[], props: Partial<ScoreProps> = {}) {
    this.parts = parts.map((part) => {
      if (part instanceof Fragment) {
        part = new Part(part);
      }
      return part;
    });
    this.ids = parts.map((_, i) => `P${i + 1}`);
    this.props = {
      ...this.props,
      ...props,
    };
  }

  static get pitchNames() {
    return Score._pitchNames;
  }

  static set pitchNames(pitchNames: string[]) {
    if (pitchNames.length !== 12) {
      throw Error("Pitch names must have 12 entries");
    }
    Score._pitchNames = pitchNames;
  }

  output(format: OutputFormats) {
    const xmlProps = { closeEmptyElement: true, tightClose: true };
    switch (format) {
      case "mei":
        return toXml(this.toMei(), xmlProps);
      case "eventData":
        return this.toEventData();
      case "beatData":
        return this.toBeatData();
      case "timingData":
        return this.toTimingData();
      default:
        return toXml(this.toMusicXml(), xmlProps);
    }
  }

  toMusicXml() {
    return u("root", [
      u("instruction", { name: "xml" }, 'version="1.0" encoding="UTF-8"'),
      u("doctype", {
        name: "score-partwise",
        public: "-//Recordare//DTD MusicXML 3.1 Partwise//EN",
        system: "http://www.musicxml.org/dtds/partwise.dtd",
      }),
      x("score-partwise", [
        x(
          "part-list",
          this.parts.map(({ props }, i) =>
            x("score-part", { id: this.ids[i] }, [x("part-name", props.name)])
          )
        ),
        ...this.parts.map((part, i) => part.toMusicXml(this.ids[i])),
      ]),
    ]);
  }

  toMei() {
    let staffDefNum = 1;
    let staffNum = 1;
    const directions: MeiDirections[] = [];
    return u("root", [
      u("instruction", { name: "xml" }, 'version="1.0" encoding="UTF-8"'),
      x("mei", [
        x("meiHead"),
        x("music", [
          x("body", [
            x("mdiv", { n: 1 }, [
              x("score", [
                x("scoreDef", [
                  x("staffGrp", [
                    this.parts.map(({ fragment, props }) =>
                      x(
                        "staffGrp",
                        {
                          label: props.name,
                          symbol: "bracket",
                          "bar.thru": "true",
                        },
                        fragment.children.map((_, i) =>
                          x("staffDef", {
                            n: staffDefNum++,
                            lines: 5,
                            ...setClef(props.clefs[i]),
                          })
                        )
                      )
                    ),
                  ]),
                ]),
                x("section", { n: 1 }, [
                  x("measure", { n: 1 }, [
                    ...this.parts.map((part) => {
                      const props = { staffNum, directions };
                      const output = part.toMei(props);
                      staffNum += part.fragment.children.length;
                      return output;
                    }),
                    ...directions.map(({ id, directions }) =>
                      directions.map((direction) => direction.toMei(id))
                    ),
                  ]),
                ]),
              ]),
            ]),
          ]),
        ]),
      ]),
    ]);
  }

  toEventData() {
    let lcd = new Fraction(1);
    for (const part of this.parts) {
      lcd = lcd.lcm(part.fragment.lcd);
    }
    const props = {
      lcd,
      timeMod: new Fraction(1),
    };
    return {
      divisions: lcd.valueOf(),
      events: this.parts.map((part) => part.toEventData(props)),
    };
  }

  toBeatData() {
    const lcd = this.parts.reduce(
      (lcd, part) => (lcd = lcd.lcm(part.fragment.lcd)),
      new Fraction(1)
    );

    return JSON.stringify(
      this.parts.reduce<{
        [index: string]: { divisions: number; beats: number[] };
      }>((state, part) => {
        state[part.name] = {
          divisions: lcd.valueOf(),
          beats: part.toBeatData(lcd),
        };

        return state;
      }, {})
    );
  }

  toTimingData() {
    const tempoData = this.parts.map((part) => part.toTimingData());
    const tempoParts: TempoParts = tempoData.map((part) => part.tempos);
    const endPositions = tempoData.map((part) => part.endPosition);
    const endPosition = endPositions
      .reduce((state, current) =>
        current.compare(state) > 0 ? current : state
      )
      .toFraction();
    const meterParts: MeterParts = endPositions.map((endPosition) => [
      {
        meter: "open",
        startPosition: "0",
        endPosition: endPosition.toFraction(),
        beatPattern: ["1/4"],
      },
    ]);
    const players: Players = this.parts.reduce<Record<string, Player>>(
      (state, part, i) => {
        state[part.name] = {
          meterPartIndex: i,
          tempoPartIndex: i,
        };

        return state;
      },
      {}
    );

    const timingData: ScoreTimingData = {
      baseTempo: this.props.baseTempo,
      endPosition,
      meterParts,
      tempoParts,
      players,
    };

    return JSON.stringify(timingData);
  }
}

export default Score;
