import { LitElement, css, html } from "lit";
import { customElement, property } from "lit/decorators.js";
import { connect } from "pwa-helpers/connect-mixin";
import { createSelector } from "@reduxjs/toolkit";
import store, { RootState } from "../../store";
import highlightEvents from "./highlightEvents";
import scroll from "./scroll";

const getScrollPos = (state: RootState) => state.scrollPos;
const getScoreEvents = (state: RootState) => state.scoreEvents;
const getScoreEventIndexes = (state: RootState) => state.scoreEventIndexes;
const getCurrentIds = createSelector(
  getScoreEvents,
  getScoreEventIndexes,
  (scoreEvents, scoreEventIndexes) =>
    scoreEvents && scoreEventIndexes
      ? scoreEventIndexes.flatMap((part, partIndex) =>
          part.map(
            (currentIndex, sequenceIndex) =>
              scoreEvents[partIndex][sequenceIndex][currentIndex].id
          )
        )
      : null
);

@customElement("score-content")
class ScoreContent extends connect(store)(LitElement) {
  @property()
  svg: SVGSVGElement | null = null;
  viewBox = { x: 0, y: 0, width: 0, height: 0 };
  @property()
  scrollPos = 0;
  @property()
  currentIds: string[] | null = null;
  previousIds: string[] | null = null;

  static get styles() {
    return css`
      ::slotted(svg) {
        flex-grow: 1;
        height: 100%;
      }
    `;
  }

  newScore() {
    this.previousIds = null;
    const svg = this.querySelector("svg");
    if (svg) {
      this.svg = svg;
      const viewBox = this.svg.getAttribute("viewBox");
      if (viewBox) {
        let [x, y, width, height] = viewBox
          .split(" ")
          .map((value) => parseInt(value));
        height = height < 1000 ? 1000 : height;
        this.svg.setAttribute("viewBox", `${x} ${y} ${width} ${height}`);
        this.svg.setAttribute("overflow", "hidden");
        this.svg.setAttribute(
          "preserveAspectRatio",
          `xMinYMin ${width > 2 * height ? "slice" : "meet"}`
        );
        this.viewBox = { x, y, width, height };
        scroll(this);
      }
    }
  }

  stateChanged(state: RootState) {
    this.currentIds = getCurrentIds(state);
    this.scrollPos = getScrollPos(state);
  }

  updated(changedProperties: any) {
    if (changedProperties.has("svg") || changedProperties.has("currentIds")) {
      highlightEvents(this);
    }
    if (changedProperties.has("scrollPos")) {
      scroll(this);
    }
  }

  render() {
    return html` <slot @slotchange=${this.newScore}></slot> `;
  }
}

export default ScoreContent;
